{-# LANGUAGE OverloadedRecordDot #-}

module Day2 where

import Control.Applicative ((<|>))
import Data.Char (isDigit)
import Data.Functor (($>))
import Data.IntMap (IntMap)
import Data.IntMap qualified as IntMap
import Data.Maybe (fromMaybe)
import Text.ParserCombinators.ReadP

run :: IO ()
run = do
  games <- readInput
  putStrLn $ "Part 1 Answer: " <> show (part1 games)
  putStrLn $ "Part 2 Answer: " <> show (part2 games)

maxRed, maxGreen, maxBlue :: Int
maxRed = 12
maxGreen = 13
maxBlue = 14

part1 :: Games -> Int
part1 = sum . IntMap.keys . IntMap.filter isCompatibleGame

isCompatibleGame :: [Revelation] -> Bool
isCompatibleGame = all isCompatibleRevelation

isCompatibleRevelation :: Revelation -> Bool
isCompatibleRevelation revelation =
  revelation.red <= maxRed
    && revelation.green <= maxGreen
    && revelation.blue <= maxBlue

part2 :: Games -> Int
part2 = sum . IntMap.map power

power :: [Revelation] -> Int
power rs =
  let redPower = maximum $ map red rs
      greenPower = maximum $ map green rs
      bluePower = maximum $ map blue rs
   in redPower * greenPower * bluePower

type Games = IntMap [Revelation]

data Revelation = Revelation {red :: Int, green :: Int, blue :: Int}

fromAssocList :: [(Colour, Int)] -> Revelation
fromAssocList m =
  -- A fold is probably faster, but its a list of max 3 elements
  Revelation
    { red = fromMaybe 0 $ lookup Red m,
      green = fromMaybe 0 $ lookup Green m,
      blue = fromMaybe 0 $ lookup Blue m
    }

data Colour = Red | Green | Blue
  deriving (Eq, Ord, Show)

readInput :: IO Games
readInput = do
  inputStr <- getContents
  case readP_to_S (parseGames <* eof) inputStr of
    [] -> error "Parser failed"
    ((g, _) : _) -> pure g

parseGameId :: ReadP Int
parseGameId = do
  _ <- string "Game "
  gameId <- read <$> many1 (satisfy isDigit)
  _ <- string ": "
  pure gameId

parseColour :: ReadP Colour
parseColour =
  do
    string "red" $> Red
    <|> (string "green" $> Green)
    <|> (string "blue" $> Blue)

parseColourRevelation :: ReadP (Colour, Int)
parseColourRevelation = do
  n <- read <$> many1 (satisfy isDigit)
  _ <- skipSpaces
  colour <- parseColour
  pure (colour, n)

parseRevelation :: ReadP Revelation
parseRevelation = do
  fromAssocList <$> sepBy1 parseColourRevelation (string ", ")

-- >>> readP_to_S (parseGame <* eof) "Game 1: 3 blue, 4 red; 1 red, 2 green, 6 blue; 2 green"
-- [((1,[fromList [(Red,4),(Blue,3)],fromList [(Red,1),(Green,2),(Blue,6)],fromList [(Green,2)]]),"")]
parseGame :: ReadP (Int, [Revelation])
parseGame = do
  gameId <- parseGameId
  revelations <- sepBy1 parseRevelation (string "; ")
  pure (gameId, revelations)

parseGames :: ReadP Games
parseGames =
  IntMap.fromList <$> sepBy1 parseGame (string "\n")
