module Day3 where

import Data.Char
import Data.Containers.ListUtils (nubOrd)
import Data.IntervalMap.FingerTree (Interval (..), IntervalMap)
import Data.IntervalMap.FingerTree qualified as IntervalMap
import Data.Map (Map)
import Data.Map.Lazy qualified as Map
import Data.Maybe (mapMaybe)
import Data.Set qualified as Set

run :: IO ()
run = do
  (parts, numbers) <- readInput
  putStrLn $ "Part 1 Answer: " <> show (part1 parts numbers)
  putStrLn $ "Part 2 Answer: " <> show (part2 parts numbers)

part1 :: PartsMap -> NumbersMap -> Int
part1 parts numbers = do
  let partCoords = Map.keysSet parts
      potentialPartNumberCoords =
        Set.unions $ Set.map (Set.fromList . neighbours) partCoords
      partNumberWithCoords =
        Set.unions $
          Set.map
            (\c -> Set.fromList $ IntervalMap.search c numbers)
            potentialPartNumberCoords
      partNumbers = map snd $ Set.toList partNumberWithCoords
  sum partNumbers

neighbours :: Coord -> [Coord]
neighbours (x, y) =
  [ (x1, y1)
    | x1 <- [x - 1 .. x + 1],
      y1 <- [y - 1 .. y + 1]
  ]

part2 :: PartsMap -> NumbersMap -> Int
part2 parts numbers =
  let starPartCoords = Map.keys $ Map.filter (== '*') parts
      gearRatios = mapMaybe (gearRatio numbers) starPartCoords
   in sum gearRatios

gearRatio :: NumbersMap -> Coord -> Maybe Int
gearRatio numbers gearCoord =
  let neighbourNums = nubOrd $ concatMap (\c -> nubOrd $ IntervalMap.search c numbers) (neighbours gearCoord)
   in case neighbourNums of
        [(_, n1), (_, n2)] -> Just (n1 * n2)
        _ -> Nothing

type Coord = (Int, Int)

type PartsMap = Map Coord Char

type NumbersMap = IntervalMap Coord Int

readInput :: IO (PartsMap, NumbersMap)
readInput = do
  ls <- lines <$> getContents
  pure $ foldr parseLine (mempty, mempty) $ zip [0 ..] ls

parseLine :: (Int, String) -> (PartsMap, NumbersMap) -> (PartsMap, NumbersMap)
parseLine (x, line) (accParts, accNumbers) =
  (parsePartsLine x line accParts, parseNumbersLine x line accNumbers)

parsePartsLine :: Int -> String -> PartsMap -> PartsMap
parsePartsLine x line acc =
  let withIndices = zip [0 ..] line
      parts = filter (\c -> snd c `notElem` "0123456789.\n") withIndices
      partsWithX = map (\(y, c) -> ((x, y), c)) parts
   in acc <> Map.fromList partsWithX

parseNumbersLine :: Int -> String -> NumbersMap -> NumbersMap
parseNumbersLine x = go 0 Nothing
  where
    go :: Int -> Maybe (Int, Int) -> [Char] -> NumbersMap -> NumbersMap
    go _ Nothing [] m = m
    go curY Nothing (c : rest) m
      | isDigit c = go (curY + 1) (Just (curY, digitToInt c)) rest m
      | otherwise = go (curY + 1) Nothing rest m
    go curY (Just (startY, curNum)) [] m =
      IntervalMap.insert (Interval (x, startY) (x, curY - 1)) curNum m
    go curY (Just (startY, curNum)) (c : rest) m
      | isDigit c = go (curY + 1) (Just (startY, curNum * 10 + digitToInt c)) rest m
      | otherwise = go (curY + 1) Nothing rest $ IntervalMap.insert (Interval (x, startY) (x, curY - 1)) curNum m
