module Day4 where

import Data.Char
import Data.IntMap (IntMap)
import Data.IntMap.Lazy qualified as IntMap
import Text.ParserCombinators.ReadP
import Data.Maybe (fromMaybe)

run :: IO ()
run = do
  cards <- readInput
  putStrLn $ "Part 1 Answer: " <> show (part1 cards)
  putStrLn $ "Part 2 Answer: " <> show (part2 cards)

part1 :: Cards -> Int
part1 cards =
  sum $ IntMap.map (uncurry points) cards

matches :: WinningNumbers -> ActualNumbers -> Int
matches winners = foldr (\x acc -> if x `elem` winners then acc + 1 else acc) 0

-- >>> points [41,48,83,86,17] [83,86,6,31,17,9,48,53]
-- 8
points :: WinningNumbers -> ActualNumbers -> Int
points winners actual =
  let m = matches winners actual
   in if m == 0
        then 0
        else 2 ^ (m - 1)

part2 :: Cards -> Int
part2 cards =
  sum $ IntMap.foldlWithKey' processCard (initCollection cards) cards

processCard :: CardCollection -> Int -> (WinningNumbers, ActualNumbers) -> CardCollection
processCard acc cardId (winners, actual) =
  let m = matches winners actual
      curCopies = fromMaybe 0 $ IntMap.lookup cardId acc
  in foldr (IntMap.adjust (+ curCopies)) acc [cardId+1 .. cardId+m]

type CardCollection = IntMap Int

initCollection :: Cards -> CardCollection
initCollection = IntMap.map (const 1)

type WinningNumbers = [Int]

type ActualNumbers = [Int]

type Cards = IntMap (WinningNumbers, ActualNumbers)

readInput :: IO Cards
readInput = do
  input <- getContents
  case readP_to_S (parseCards <* eof) input of
    [] -> error "parser failed"
    [(cards, "")] -> pure cards
    x -> error $ "parser failed: " <> show x

parseNumber :: ReadP Int
parseNumber = read <$> many1 (satisfy isDigit)

parseTwoDigitNumber :: ReadP Int
parseTwoDigitNumber = do
  digit1Chr <- satisfy (`elem` " 0123456789")
  digit2 <- digitToInt <$> satisfy isDigit
  if digit1Chr == ' '
    then pure digit2
    else pure (digitToInt digit1Chr * 10 + digit2)

-- >>> readP_to_S (parseCardId <* eof) "Card    1: "
-- [(1,"")]
parseCardId :: ReadP Int
parseCardId = do
  _ <- string "Card"
  _ <- skipSpaces
  cardId <- parseNumber
  _ <- string ": "
  pure cardId

-- >>> readP_to_S (parseNumbers <* eof) "41 48 83 86 17"
-- [([41,48,83,86,17],"")]
parseNumbers :: ReadP [Int]
parseNumbers = do
  sepBy1 parseTwoDigitNumber (char ' ')

-- >>> readP_to_S (parseCard <* eof) "Card 1: 41 48 83 86 17 | 83 86  6 31 17  9 48 53"
-- [((1,([41,48,83,86,17],[83,86,6,31,17,9,48,53])),"")]
parseCard :: ReadP (Int, (WinningNumbers, ActualNumbers))
parseCard = do
  cardId <- parseCardId
  winning <- parseNumbers
  _ <- string " | "
  actual <- parseNumbers
  pure (cardId, (winning, actual))

parseCards :: ReadP Cards
parseCards =
  IntMap.fromList <$> sepBy1 parseCard (char '\n')
