{-# LANGUAGE RecordWildCards #-}
{-# OPTIONS_GHC -Wwarn -Wall -Wunused-top-binds #-}

module Day5 (run) where

import Data.Char
import Data.IntervalMap.FingerTree (Interval (Interval), IntervalMap, high, low)
import Data.IntervalMap.FingerTree qualified as IntervalMap
import Data.Maybe (maybeToList)
import Text.ParserCombinators.ReadP

run :: IO ()
run = do
  almanac <- readInput
  putStrLn $ "Part 1 Answer: " <> show (part1 almanac)
  putStrLn $ "Part 2 Answer: " <> show (part2 almanac)

part1 :: Almanac -> Int
part1 almanac =
  minimum $ map (seedToLocation almanac) (seeds almanac)

part2 :: Almanac -> Int
part2 almanac =
  minimum $ seedToLocationMap almanac

type Seeds = [Int]

type RangeMap = IntervalMap Int Int

data Almanac = Almanac
  { seeds :: Seeds,
    seedToSoilMap :: RangeMap,
    soilToFertilizerMap :: RangeMap,
    fertilizerToWaterMap :: RangeMap,
    waterToLightMap :: RangeMap,
    lightToTemperatureMap :: RangeMap,
    temperatureToHumidityMap :: RangeMap,
    humidityToLocationMap :: RangeMap
  }
  deriving (Show)

rangeMapLookup :: (Ord a, Enum a, Enum b) => (a -> b) -> a -> IntervalMap a b -> b
rangeMapLookup f src m =
  case IntervalMap.search src m of
    [] -> f src
    -- Ignores the fact that there can be many matches as it is not defined in
    -- the problem statement
    ((srcRange, destStart) : _) -> rangeLookup srcRange destStart src

rangeLookup :: (Enum a, Enum b) => Interval a -> b -> a -> b
rangeLookup ia bStart a =
  toEnum $ fromEnum bStart + (fromEnum a - fromEnum (low ia))

seedToLocation :: Almanac -> Int -> Int
seedToLocation Almanac {..} seed =
  foldl
    (rangeMapLookup id)
    seed
    [ seedToSoilMap,
      soilToFertilizerMap,
      fertilizerToWaterMap,
      waterToLightMap,
      lightToTemperatureMap,
      temperatureToHumidityMap,
      humidityToLocationMap
    ]

seedToLocationMap :: Almanac -> RangeMap
seedToLocationMap Almanac {..} =
  foldl
    (composeRangeMaps id id)
    (seedToSeedMap seeds)
    [ seedToSoilMap,
      soilToFertilizerMap,
      fertilizerToWaterMap,
      waterToLightMap,
      lightToTemperatureMap,
      temperatureToHumidityMap,
      humidityToLocationMap
    ]
  where
    seedToSeedMap :: Seeds -> RangeMap
    seedToSeedMap [] = mempty
    seedToSeedMap [_] = mempty
    seedToSeedMap (x : y : rest) = IntervalMap.insert (Interval x (x + y - 1)) x $ seedToSeedMap rest

-- Example:
--
-- ----------------------- :: -------------------- :: -------------------- :: ------------------- :: -------------------------------------------
--        Name             ::   source interval    :: overlapping intervals::   resolution        ::  Remarks
-- ----------------------- :: -------------------- :: -------------------- :: ------------------- :: -------------------------------------------
-- 1. Perfect overlap      ::  [0, 10] -> [10, 20] :: [10, 20] -> [30, 40] :: [0, 10] -> [30, 40] ::
-- ----------------------- :: -------------------- :: ---------------------:: ------------------- :: -------------------------------------------
-- 2. Right skewed overlap ::  [0, 10] -> [15, 25] :: [10, 20] -> [30, 40] :: [0,  5] -> [35, 40] ::
--                         ::                      ::                      :: [6, 10] -> [21, 25] ::
-- ----------------------- :: -------------------- :: ---------------------:: ------------------- :: -------------------------------------------
-- 3. Left skewed overlap  ::  [0, 10] -> [ 5, 15] :: [10, 20] -> [30, 40] :: [0,  4] -> [ 5,  9] ::
--                         ::                      ::                      :: [5, 10] -> [30, 35] ::
-- ----------------------- :: -------------------- :: ---------------------:: ------------------- :: -------------------------------------------
-- 4. Containing overlap   ::  [0, 10] -> [10, 20] :: [12, 17] -> [32, 37] :: [0,  1] -> [10, 11] :: left skew
--                         ::                      ::                      :: [2,  7] -> [12, 17] :: matches
--                         ::                      ::                      :: [8, 10] -> [ 8, 10] :: right skew
-- ----------------------- :: -------------------- :: ---------------------:: ------------------- :: -------------------------------------------
-- 5. Contained overlap    ::  [0, 10] -> [10, 20] :: [ 9, 22] -> [29, 42] :: [0, 10] -> [30, 40] ::
-- ----------------------- :: -------------------- :: ---------------------:: ------------------- :: -------------------------------------------
-- 6. Multiple overlaps    ::  [0, 10] -> [10, 20] :: [ 9, 13] -> [29, 33] :: [0,  3] -> [30, 33] ::
--                         ::                      :: [15, 17] -> [35, 37] :: [4,  4] -> [14, 14] ::
--                         ::                      :: [19, 25] -> [39, 45] :: [5,  7] -> [35, 37] ::
--                         ::                      ::                      :: [8,  8] -> [18, 18] ::
--                         ::                      ::                      :: [9, 10] -> [39, 40] ::
-- ----------------------- :: -------------------- :: -------------------- :: ------------------- :: -------------------------------------------

-- >>> i = Interval
-- >>> assertEqual a b = if a == b then True else error $ "Expected: " <> show a <> "\n  Actual: " <> show b
--
-- Perfect overlap
-- >>> assertEqual Nothing $ preMatchInterval (i 0 10) 10 (i 10 20)
-- True
--
-- Right skewed overlap
-- >>> assertEqual Nothing $ preMatchInterval (i 0 10) 15 (i 10 20)
-- True
--
-- Left skewed overlap
-- >>> assertEqual (Just $ i 0 4) $ preMatchInterval (i 0 10) 5 (i 10 20)
-- True
--
-- Containing overlap
-- >>> assertEqual (Just $ i 0 1) $ preMatchInterval (i 0 10) 10 (i 12 17)
-- True
--
-- Multiple overlaps 1:
-- >>> assertEqual Nothing $ preMatchInterval (i 0 10) 10 (i 9 13)
-- True
--
-- Extra case, probably not going to happen:
-- >>> assertEqual (Just $ i 0 10) $ preMatchInterval (i 0 10) 10 (i 30 40)
-- True
preMatchInterval :: (Ord b, Enum b, Enum a, Ord a) => Interval a -> b -> Interval b -> Maybe (Interval a)
preMatchInterval ia bStart ib
  | bStart < low ib =
      Just $ Interval (low ia) (min (high ia) (succByDiff (low ia) (pred $ low ib) bStart))
  | otherwise = Nothing

-- >>> i = Interval
-- >>> assertEqual a b = if a == b then True else error $ "Expected: " <> show a <> "\n  Actual: " <> show b
--
-- Perfect overlap
-- >>> assertEqual (Just $ i 0 10) $ matchInterval (i 0 10) 10 (i 10 20)
-- True
--
-- Right skewed overlap
-- >>> assertEqual (Just $ i 0 5) $ matchInterval (i 0 10) 15 (i 10 20)
-- True
--
-- Left skewed overlap
-- >>> assertEqual (Just $ i 5 10) $ matchInterval (i 0 10) 5 (i 10 20)
-- True
--
-- Containing overlap
-- >>> assertEqual (Just $ i 2 7) $ matchInterval (i 0 10) 10 (i 12 17)
-- True
--
-- Multiple overlaps 1:
-- >>> assertEqual (Just $ i 0 3) $ matchInterval (i 0 10) 10 (i 9 13)
-- True
--
-- Extra case, probably not going to happen:
-- >>> assertEqual Nothing $ matchInterval (i 0 10) 10 (i 30 40)
-- True
matchInterval :: (Ord a, Ord b, Enum a, Enum b) => Interval a -> b -> Interval b -> Maybe (Interval a)
matchInterval ia bStart ib
  | bStart `elemInterval` ib
      || bEnd `elemInterval` ib
      || (bStart <= low ib && bEnd >= high ib) =
      Just $ Interval aStart aEnd
  | otherwise = Nothing
  where
    bEnd = succByDiff bStart (high ia) (low ia)
    aStart =
      if bStart `elemInterval` ib
        then low ia
        else succByDiff (low ia) (low ib) bStart
    aEnd =
      if bEnd `elemInterval` ib
        then high ia
        else succByDiff (low ia) (high ib) bStart

-- >>> i = Interval
-- >>> assertEqual a b = if a == b then True else error $ "Expected: " <> show a <> "\n  Actual: " <> show b
--
-- Perfect overlap
-- >>> assertEqual Nothing $ postMatchInterval (i 0 10) 10 (i 10 20)
-- True
--
-- Right skewed overlap
-- >>> assertEqual (Just $ i 6 10) $ postMatchInterval (i 0 10) 15 (i 10 20)
-- True
--
-- Left skewed overlap
-- >>> assertEqual Nothing $ postMatchInterval (i 0 10) 5 (i 10 20)
-- True
--
-- Containing overlap
-- >>> assertEqual (Just $ i 8 10) $ postMatchInterval (i 0 10) 10 (i 12 17)
-- True
--
-- Multiple overlaps 1:
-- >>> assertEqual (Just $ i 4 10) $ postMatchInterval (i 0 10) 10 (i 9 13)
-- True
--
-- Extra case, probably not going to happen:
-- >>> assertEqual Nothing $ postMatchInterval (i 0 10) 10 (i 30 40)
-- True
postMatchInterval :: (Ord a, Ord b, Enum a, Enum b) => Interval a -> b -> Interval b -> Maybe (Interval a)
postMatchInterval ia bStart ib
  | bEnd > high ib =
      Just $
        Interval
          (succByDiff (low ia) (succ $ high ib) bStart)
          (high ia)
  | otherwise = Nothing
  where
    bEnd = succByDiff bStart (high ia) (low ia)

-- >>> i = Interval
-- >>> assertEqual a b = if a == b then True else error $ "Expected: " <> show a <> "\n  Actual: " <> show b
--
-- Perfect overlap
-- >>> assertEqual [i 0 10] $ newIntervals (i 0 10) 10 [(i 10 20)]
-- True
--
-- Right skewed overlap
-- >>> assertEqual [i 0 5, i 6 10] $ newIntervals (i 0 10) 15 [(i 10 20)]
-- True
--
-- Left skewed overlap
-- >>> assertEqual [i 0 4, i 5 10] $ newIntervals (i 0 10) 5 [(i 10 20)]
-- True
--
-- Containing overlap
-- >>> assertEqual [i 0 1, i 2 7, i 8 10] $ newIntervals (i 0 10) 10 [(i 12 17)]
-- True
--
-- Multiple overlaps 1:
-- >>> assertEqual [i 0 3, i 4 4, i 5 7, i 8 8, i 9 10] $ newIntervals (i 0 10) 10 [i 9 13, i 15 17, i 19 25]
-- True
--
-- >>> assertEqual [i 0 10] $ newIntervals (i 0 10) 10 []
-- True
--
-- Extra case, probably not going to happen:
-- >>> assertEqual [i 0 10] $ newIntervals (i 0 10) 10 [(i 30 40)]
-- True
newIntervals :: (Ord a, Ord b, Enum a, Enum b) => Interval a -> b -> [Interval b] -> [Interval a]
newIntervals ia _ [] = [ia]
newIntervals ia bStart (ib : rest) =
  let pres = maybeToList $ preMatchInterval ia bStart ib
      matches = maybeToList $ matchInterval ia bStart ib
   in case postMatchInterval ia bStart ib of
        Nothing -> pres <> matches
        Just post -> pres <> matches <> newIntervals post (succByDiff bStart (low post) (low ia)) rest

composeRangeMaps :: (Ord a, Ord b, Ord c, Enum a, Enum b, Enum c, Bounded a) => (a -> b) -> (b -> c) -> IntervalMap a b -> IntervalMap b c -> IntervalMap a c
composeRangeMaps a2b b2c r1 r2 =
  let allIntervalsWithStarts = IntervalMap.intersections (Interval minBound maxBound) r1
      relevantIntervals =
        foldMap
          ( \(ia, bStart) ->
              let bEnd = succByDiff bStart (high ia) (low ia)
               in newIntervals ia bStart $
                    map fst $
                      IntervalMap.intersections (Interval bStart bEnd) r2
          )
          allIntervalsWithStarts
   in foldMap (\ia -> IntervalMap.singleton ia (rangeMapLookup b2c (rangeMapLookup a2b (low ia) r1) r2)) relevantIntervals

-- >>> enumDiff 15 10
-- 5
enumDiff :: (Enum a) => a -> a -> Int
enumDiff x y = fromEnum x - fromEnum y

-- >>> succByDiff 0 15 10
-- 5
succByDiff :: (Enum a, Enum b) => a -> b -> b -> a
succByDiff x y1 y2 = toEnum (fromEnum x + enumDiff y1 y2)

elemInterval :: (Ord a) => a -> Interval a -> Bool
a `elemInterval` i = low i <= a && a <= high i

-- * Parsing

emptyAlmanac :: Almanac
emptyAlmanac =
  Almanac
    { seeds = mempty,
      seedToSoilMap = mempty,
      soilToFertilizerMap = mempty,
      fertilizerToWaterMap = mempty,
      waterToLightMap = mempty,
      lightToTemperatureMap = mempty,
      temperatureToHumidityMap = mempty,
      humidityToLocationMap = mempty
    }

readInput :: IO Almanac
readInput = do
  input <- getContents
  case readP_to_S (almanacParser <* eof) input of
    [] -> error "Parser failed"
    [(almanac, "")] -> pure almanac
    xs -> error $ "Parser failed: " <> show xs

-- >>> readP_to_S numberParser "123"
-- [(123,"")]
numberParser :: ReadP Int
numberParser =
  read <$> munch1 isDigit

-- >>> readP_to_S (seedsParser <* eof) "seeds: 79 14 55 13"
-- [([79,14,55,13],"")]
seedsParser :: ReadP Seeds
seedsParser = do
  _ <- string "seeds: "
  sepBy1 numberParser (char ' ')

-- >>> readP_to_S (rangeParser <* eof) "52 50 48"
-- [((52,50,48),"")]
rangeParser :: ReadP (Int, Int, Int)
rangeParser = do
  destRangeStart <- numberParser
  _ <- char ' '
  srcRangeStart <- numberParser
  _ <- char ' '
  rangeLen <- numberParser
  pure (destRangeStart, srcRangeStart, rangeLen)

-- >>> readP_to_S (rangeMapParser <* eof) "seed-to-soil map:\n50 98 2\n52 50 48"
-- [(("seed-to-soil",insert (Interval 50 97) 52 $ insert (Interval 98 99) 50 $ empty),"")]
rangeMapParser :: ReadP (String, RangeMap)
rangeMapParser = do
  mapName <- munch1 (/= ' ')
  _ <- string " map:\n"
  rangeMap <-
    foldr
      ( \(destStart, srcStart, len) ->
          IntervalMap.insert (Interval srcStart (srcStart + len - 1)) destStart
      )
      mempty
      <$> sepBy1 rangeParser (char '\n')
  pure (mapName, rangeMap)

almanacParser :: ReadP Almanac
almanacParser = do
  parsedSeeds <- seedsParser
  _ <- string "\n\n"
  rangeMaps <- sepBy1 rangeMapParser (string "\n\n")
  pure $ foldr (uncurry updateAlmanacByName) (emptyAlmanac {seeds = parsedSeeds}) rangeMaps

updateAlmanacByName :: String -> RangeMap -> Almanac -> Almanac
updateAlmanacByName name rangeMap almanac =
  case name of
    "seed-to-soil" -> almanac {seedToSoilMap = rangeMap}
    "soil-to-fertilizer" -> almanac {soilToFertilizerMap = rangeMap}
    "fertilizer-to-water" -> almanac {fertilizerToWaterMap = rangeMap}
    "water-to-light" -> almanac {waterToLightMap = rangeMap}
    "light-to-temperature" -> almanac {lightToTemperatureMap = rangeMap}
    "temperature-to-humidity" -> almanac {temperatureToHumidityMap = rangeMap}
    "humidity-to-location" -> almanac {humidityToLocationMap = rangeMap}
    _ -> almanac -- Ignore the wrong input ¯\_(ツ)_/¯
