use std::io;
use std::io::Read;
use std::cmp::Ordering;

pub fn run() {
    let mut lines : Vec<u8> = Vec::new();
    let read_result = io::stdin().read_to_end(&mut lines);
    match read_result {
        Err(e) => println!("Error while reading input: {}", e),
        Ok(_) => {
            match part1(lines.clone()) {
                Some(s) => println!("Part 1 Answer: {}", s),
                None => println!("Part 1 Answer: failed to compute")
            };
            match part2(lines) {
                Some(s) => println!("Part 2 Answer: {}", s),
                None => println!("Part 2 Answer: failed to compute")
            }
        }
    }
}

fn part1(lines: Vec<u8>) -> Option<u32> {
    general_solution(lines, first_digit, last_digit)
}

fn part2(lines: Vec<u8>) -> Option<u32> {
    general_solution(lines, first_digit_with_creativity, last_digit_with_creativity)
}

fn general_solution(data: Vec<u8>, parse_tens: fn(&str) -> Option<u32>,parse_ones: fn(&str) -> Option<u32>) -> Option<u32> {
    // Too hard to carry around the Result. Using `ok()` to convert it to
    // `Option` doesn't work for some reason.
    let text = String::from_utf8(data).unwrap();
    let lines = text.lines();
    let calibration_values : Option<Vec<u32>> =
        lines.map(|line| {
            let tens = parse_tens(line);
            let ones = parse_ones(line);
            tens.and_then(
                |t| ones.and_then(
                    |o| Some(t * 10 + o)
                )
            )
        }).collect();
    calibration_values.and_then(|x| Some(x.into_iter().sum()))
}

fn first_digit(s: &str) -> Option<u32> {
    for c in s.chars() {
        if c.is_digit(10) {
            return c.to_digit(10);
        }
    }
    None
}

fn last_digit(s: &str) -> Option<u32> {
    for c in s.chars().rev() {
        if c.is_digit(10) {
            return c.to_digit(10);
        }
    }
    None
}

struct SearchData {
    value : u32,
    position : Option<usize>,
}

fn lower_search_data(d1: &SearchData, d2: &SearchData) -> Ordering {
    match (d1.position, d2.position) {
        (None, None) => Ordering::Equal,
        (Some(_), None) => Ordering::Less,
        (None, Some(_)) => Ordering::Greater,
        (Some(x), Some(y)) => x.cmp(&y)
    }
}

fn greater_search_data(d1: &SearchData, d2: &SearchData) -> Ordering {
    match (d1.position, d2.position) {
        (None, None) => Ordering::Equal,
        (Some(_), None) => Ordering::Greater,
        (None, Some(_)) => Ordering::Less,
        (Some(x), Some(y)) => x.cmp(&y)
    }
}

fn first_digit_with_creativity(s: &str) -> Option<u32> {
    let search_results = [ SearchData{ value: 0, position: s.find("0")},
                           SearchData{ value: 1, position: s.find("1")},
                           SearchData{ value: 2, position: s.find("2")},
                           SearchData{ value: 3, position: s.find("3")},
                           SearchData{ value: 4, position: s.find("4")},
                           SearchData{ value: 5, position: s.find("5")},
                           SearchData{ value: 6, position: s.find("6")},
                           SearchData{ value: 7, position: s.find("7")},
                           SearchData{ value: 8, position: s.find("8")},
                           SearchData{ value: 9, position: s.find("9")},
                           SearchData{ value: 0, position: s.find("zero")},
                           SearchData{ value: 1, position: s.find("one")},
                           SearchData{ value: 2, position: s.find("two")},
                           SearchData{ value: 3, position: s.find("three")},
                           SearchData{ value: 4, position: s.find("four")},
                           SearchData{ value: 5, position: s.find("five")},
                           SearchData{ value: 6, position: s.find("six")},
                           SearchData{ value: 7, position: s.find("seven")},
                           SearchData{ value: 8, position: s.find("eight")},
                           SearchData{ value: 9, position: s.find("nine")}
    ];

    search_results
        .into_iter()
        .min_by(lower_search_data)
        .and_then(|x| x.position.and_then(|_| Some(x.value)))
}

fn last_digit_with_creativity(s: &str) -> Option<u32> {
    let search_results = [ SearchData{ value: 0, position: s.rfind("0")},
                           SearchData{ value: 1, position: s.rfind("1")},
                           SearchData{ value: 2, position: s.rfind("2")},
                           SearchData{ value: 3, position: s.rfind("3")},
                           SearchData{ value: 4, position: s.rfind("4")},
                           SearchData{ value: 5, position: s.rfind("5")},
                           SearchData{ value: 6, position: s.rfind("6")},
                           SearchData{ value: 7, position: s.rfind("7")},
                           SearchData{ value: 8, position: s.rfind("8")},
                           SearchData{ value: 9, position: s.rfind("9")},
                           SearchData{ value: 0, position: s.rfind("zero")},
                           SearchData{ value: 1, position: s.rfind("one")},
                           SearchData{ value: 2, position: s.rfind("two")},
                           SearchData{ value: 3, position: s.rfind("three")},
                           SearchData{ value: 4, position: s.rfind("four")},
                           SearchData{ value: 5, position: s.rfind("five")},
                           SearchData{ value: 6, position: s.rfind("six")},
                           SearchData{ value: 7, position: s.rfind("seven")},
                           SearchData{ value: 8, position: s.rfind("eight")},
                           SearchData{ value: 9, position: s.rfind("nine")}
    ];

    search_results
        .into_iter()
        .max_by(greater_search_data)
        .and_then(|x| x.position.and_then(|_| Some(x.value)))
}
